module ControlCompute
export PID, compute!, reset!, Integral, Derivative, Proportional

struct Savgol
    filter::Array{Float64, 2}
    poly_order::Int8
    sample_time::Float64
    derivative::Int8
    convoluate::Bool
    function Savgol(size::Int64, poly_order::Int64, deriv::Int64=0, delta::Float64=1.0, conv::Bool=false)
        half_size, rem = divrem(size, 2)
        if rem == 0
            throw(ArgumentError("size must be odd."))
        end
        M = [-half_size:half_size;] .^ [0:poly_order;]';
        y = zeros(poly_order+1)';
        y[deriv+1] = factorial(deriv) / delta^deriv;
        scal = y*inv(M'*M)*M'
        if conv
            scal = scal[end:-1:1]
        end
        new(scal, poly_order, delta, conv)
    end
end

abstract type Block end

mutable struct Derivative <: Block
    k::Float64
    last_points::Array{Float64, 1}
    filter::Savgol
    last_time::Float64
    refresh_time::Float64
    output::Float64
    Derivative(k::Float64=1.0, refresh_time::Float64=1.0, size::Int64=3, poly_order::Int64=1) = new(k, zeros(size), Savgol(size, poly_order, 1, refresh_time), 0.0, refresh_time, 0.0)
end

mutable struct Proportional <: Block
    k::Float64
    last_time::Float64
    refresh_time::Float64
    output::Float64
    Proportional(k=1.0, refresh_time=1.0) = new(k, 0.0, refresh_time, 0.0)
end

abstract type AbstractSaturate <: Block end

struct Saturate <: AbstractSaturate
    min::Union{Float64, Nothing}
    max::Union{Float64, Nothing}
    output::Float64
    last_time::Float64
    refresh_time::Float64
    Saturate(min=nothing, max=nothing) = new(min, max, 0.0, 0.0, 1.0)
end

mutable struct Integral <: AbstractSaturate
    k::Float64
    last_time::Float64
    refresh_time::Float64
    min::Union{Float64, Nothing}
    max::Union{Float64, Nothing}
    output::Float64
    Integral(k=1.0, refresh_time=1.0, min=nothing, max=nothing) = new(k, 0.0, refresh_time, min, max, 0.0)
end

mutable struct PID <: Block
    P::Proportional
    I::Integral
    D::Derivative
    last_time::Float64
    refresh_time::Float64
    output::Float64
    PID(kp=0.0, ki=0.0, kd=0.0, refresh_time=1.0) = new(Proportional(kp, refresh_time), Integral(ki, refresh_time), Derivative(kd, refresh_time), 0.0, refresh_time, 0.0)
end

mutable struct Differenciate <: Block
    last_time::Float64
    refresh_time::Float64
    output::Float64
    Differenciate() = new(time(), 0.0, 0.0)
end

"""
    reset!(block::Block)

Reset the `last_time` and `output` attribute of the block.
"""
function reset!(block::Block, output::Float64=0.0)
    block.last_time = time()
    block.output = output
end


"""
    check_time(block::Block, delta_time::Float64=0)

Check the time elapsed since the last computation. Returns time_elapsed >= block.refresh_time

The time in seconds elapsed since the last computation is given by `delta_time`.
If it equals 0.0, the delta is computed using `time()`.
"""
function check_time(block::Block, delta_time::Float64=0)
    t = time()
    if delta_time == 0
        delta_time = t - block.last_time
    end
    return delta_time >= block.refresh_time, delta_time
end

"""
    compute!(pid::PID, input::Float64, delta_time::Float64=0.0, measure::Float64=nothing)

Compute the output of a PID corrector using the given input.

Uses the `check_time` function. If not enough time has passed since the last
compute, it just keeps the last computed output. If you want to compute the
derivative on the measurement rather than on the error to avoid derrivative jump,
use the `measure` parameter.
"""
function compute!(pid::PID, input::Float64, delta_time::Float64=0.0, measure::Union{Float64, Nothing}=nothing)
    if ! check_time(pid, delta_time)
        return
    end
    compute!(pid.P, input, delta_time)
    compute!(pid.I, input, delta_time)
    if measure == nothing
        compute!(pid.D, input, delta_time)
    else
        compute!(pid.D, measure, delta_time)
    end

    pid.last_time = time();
    pid.output = pid.P.output + pid.I.output + pid.D.output
end

"""
    compute!(block::Proportional, input::Float64, delta_time::Float64=0.0)

Compute the output of a Proportional block using the given input.

Uses the `check_time` function. If not enough time has passed since the last
compute, it just keeps the last computed output.
"""
function compute!(block::Proportional, input::Float64, delta_time::Float64=0.0)
    if ! check_time(block, delta_time)
        return
    end

    block.last_time = time();
    block.output = block.k * input
end

"""
    compute!(block::Saturate, input::Float64, delta_time::Float64=0.0)

Compute the output of a Saturate block using the given input.

Uses the `check_time` function. If not enough time has passed since the last
compute, it just keeps the last computed output.
"""
function compute!(block::Saturate, input::Float64, delta_time::Float64=0.0)
    if ! check_time(block, delta_time)
        return
    end

    block.output = input
    if block.min != nothing
        block.output = max(block.output, block.min)
    end
    if block.max != nothing
        block.output = min(block.output, block.max)
    end

    block.last_time = time();
    block.output
end

"""
    compute!(block::Integral, input::Float64, delta_time::Float64=0.0)

Compute the output of a Integral block using the given input.

Uses the `check_time` function. If not enough time has passed since the last
compute, it just keeps the last computed output.
"""
function compute!(block::Integral, input::Float64, delta_time::Float64=0.0)
    delta_ok, delta = check_time(block, delta_time)
    if ! delta_ok
        return
    end

    block.output += block.k * delta * input
    println(block.output)
    if block.min != nothing
        block.output = max(block.output, block.min)
    end
    if block.max != nothing
        block.output = min(block.output, block.max)
    end

    block.last_time = time();
    block.output
end

"""
    compute!(block::Derivative, input::Float64, delta_time::Float64=0.0)

Compute the output of a Derivative block using the given input.

Uses the `check_time` function. If not enough time has passed since the last
compute, it just keeps the last computed output.
"""
function compute!(block::Derivative, input::Float64, delta_time::Float64=0.0)
    if ! check_time(block, delta_time)
        return
    end

    block.last_points = [input, block.last_points[1:end-1]...];
    block.output = block.k * (block.filter.filter * block.last_points)[1];

    block.last_time = time();
    block.output
end

"""
    compute!(block::Differenciate, input_plus::Float64, input_minus:Float64, delta_time::Float64=0.0)

Compute the output of a Differenciate block, i.e. `input_plus - input_minus`.

Uses the `check_time` function. If not enough time has passed since the last
compute, it just keeps the last computed output.
"""
function compute!(block::Differenciate, input_plus::Float64, input_minus::Float64, delta_time::Float64=0.0)
    if ! check_time(block, delta_time)
        return
    end

    block.last_time = time();
    block.output = input_plus - input_minus
end

end  # module ControlCompute
