#!/usr/bin/env julia

doc = """Control Node.

Allow to create ROS control node among pid, proportional, integral, derivative,
input, differenciate and saturate (see the `--node` option).

Usage:
  control_node.jl --node=<node> [--name=<name>] [<repeating-argument>...]
  control_node.jl -h | --help

Options:
  -h --help     Show this screen.
  --node=<node> Specify the node.
  --name=<name> Set the name of the node.

"""

push!(LOAD_PATH,"/home/klafyvel/Documents/supelec/2A/projet/workspace/src/detect_targets/scripts/")
ENV["PYTHON"] = "/usr/bin/python"
using PyCall

using RobotOS
@rosimport std_msgs.msg: Float64
rostypegen()

using DocOpt

server_module = pyimport("dynamic_reconfigure.server");

ProportionalNodeConfig = pyimport("detect_targets.cfg.ProportionalNodeConfig")
IntegralNodeConfig = pyimport("detect_targets.cfg.IntegralNodeConfig")
DerivativeNodeConfig = pyimport("detect_targets.cfg.DerivativeNodeConfig")
PIDNodeConfig = pyimport("detect_targets.cfg.PIDNodeConfig")
InputNodeConfig = pyimport("detect_targets.cfg.InputNodeConfig")
SaturateNodeConfig = pyimport("detect_targets.cfg.SaturateNodeConfig")

include("control_compute.jl");
# using ControlCompute

mutable struct ControlNode{T<:ControlCompute.Block}
    inputs::Dict{String, Float64}
    input_topics::Dict{String, Subscriber{std_msgs.msg.Float64Msg}}
    output_topic::Publisher{std_msgs.msg.Float64Msg}
    block::T
end

function initNode(block::T) where {T<:ControlCompute.Block}
    inputs = Dict{String, Float64}("input"=>0.0)
    input_topics = Dict{String, Subscriber{std_msgs.msg.Float64Msg}}()
    output_topic = Publisher{std_msgs.msg.Float64Msg}("output", queue_size=1)
    node = ControlNode(inputs, input_topics, output_topic, block)
    get!(
        node.input_topics,
        "input",
        Subscriber{std_msgs.msg.Float64Msg}("input", create_output, (node,))
    )
    # get!(
    #     node.input_topics,
    #     "reset",
    #     Subscriber{std_msgs.msg.Float64Msg}("reset", reset_node, (node,))
    # )
    node
end

function partial(f,a...)
  (b...) -> f(a...,b...)
end

function ControlNode(block::ControlCompute.PID)
    node = initNode(block)
    callback = partial(on_reconf, node)
    server = server_module.Server(PIDNodeConfig, callback)
    node.input_topics["measure"] = Subscriber{std_msgs.msg.Float64Msg}("measure", create_output_diff, (node,))
    node, server
end

function ControlNode(block::ControlCompute.Proportional)
    node = initNode(block)
    callback = partial(on_reconf, node)
    server = server_module.Server(ProportionalNodeConfig, callback)
    node, server
end

function ControlNode(block::ControlCompute.Integral)
    node = initNode(block)
    callback = partial(on_reconf, node)
    server = server_module.Server(IntegralNodeConfig, callback)
    node, server
end

function ControlNode(block::ControlCompute.Derivative)
    node = initNode(block)
    callback = partial(on_reconf, node)
    server = server_module.Server(DerivativeNodeConfig, callback)
    node, server
end

function ControlNode(block::ControlCompute.Saturate)
    node = initNode(block)
    callback = partial(on_reconf, node)
    server = server_module.Server(SaturateNodeConfig, callback)
    node, server
end

function ControlNode(block::ControlCompute.Differenciate)
    node = initNode(block)
    node.input_topics["measure"] = Subscriber{std_msgs.msg.Float64Msg}("measure", create_output_diff, (node,))
    node, nothing
end

function create_output(input, node::ControlNode{T}) where {T<:ControlCompute.Block}
    node.inputs["input"] = input
    ControlCompute.compute!(node, input)
    publish(node.output_topic, node.output)
end

function create_output(input, node::ControlNode{ControlCompute.PID})
    node.inputs["input"] = input
    ControlCompute.compute!(node, input, 0.0, node.inputs["measure"])
    publish(node.output_topic, node.output)
end

function create_output(input, node::ControlNode{ControlCompute.Differenciate})
    node.inputs["input"] = input
    ControlCompute.compute!(node, input, 0.0, node.inputs["measure"])
    publish(node.output_topic, node.output)
end

function create_output_diff(input_diff, node::ControlNode{ControlCompute.PID})
    input = node.inputs["input"]
    node.inputs["measure"] = input_diff
    ControlCompute.compute!(node, input, input_diff)
    publish(node.output_topic, node.output)
end

function create_output_diff(input_diff, node::ControlNode{ControlCompute.Differenciate})
    input = node.inputs["input"]
    node.inputs["measure"] = input_diff
    ControlCompute.compute!(node, input, 0.0, input_diff)
    publish(node.output_topic, node.output)
end

function on_reconf(node::ControlNode{ControlCompute.PID}, config, status)
    refresh_time = config["refresh_time"]
    node.block.P = ControlCompute.Proportional(
        config["kp"],
        refresh_time
    )
    node.block.I = ControlCompute.Integral(
        config["ki"],
        refresh_time,
        config["use_min"] ? config["min"] : nothing,
        config["use_max"] ? config["max"] : nothing
    )
    node.block.D = ControlCompute.Derivative(
        config["dom"] ? -config["kd"] : config["kd"],
        refresh_time,
        config["size"],
        config["poly_order"]
    )
    node.block.refresh_time = refresh_time
    config
end

function on_reconf(node::ControlNode{ControlCompute.Proportional}, config, status)
    refresh_time = config["refresh_time"]
    node.block.k = config["k"]
    node.block.refresh_time = refresh_time
    config
end

function on_reconf(node::ControlNode{ControlCompute.Saturate}, config, status)
    refresh_time = config["refresh_time"]
    node.block.min = config["use_min"] ? config["min"] : nothing
    node.block.max = config["use_max"] ? config["max"] : nothing
    node.block.refresh_time = refresh_time
    config
end

function on_reconf(node::ControlNode{ControlCompute.Integral}, config, status)
    refresh_time = config["refresh_time"]
    node.block.k = config["k"]
    node.block.min = config["use_min"] ? config["min"] : nothing
    node.block.max = config["use_max"] ? config["max"] : nothing
    node.block.refresh_time = refresh_time
    config
end

function on_reconf(node::ControlNode{ControlCompute.Derivative}, config, status)
    refresh_time = config["refresh_time"]
    node.block.k = config["k"]
    node.block.filter = ControlCompute.Savgol(
        config["size"],
        config["poly_order"],
        1,
        refresh_time
    )
    node.block.refresh_time = refresh_time
    config
end

mutable struct InputNode
    input::Float64
    output_topic::Union{Nothing, Publisher{std_msgs.msg.Float64Msg}}
    function InputNode()
        node = new(0, nothing)
        node.output_topic = Publisher{std_msgs.msg.Float64Msg}("output", queue_size=1)
        #<PyCall.jlwrap getfield(Main, Symbol("##3#4")){typeof(on_reconf_input),Tuple{InputNode}}(on_reconf_input, (InputNode(0.0, Publisher{Main.std_msgs.msg.Float64Msg}(PyObject <rospy.topics.Publisher object at 0x7f15f7e97710>)),))>
        #<PyCall.jlwrap PyCall.FuncWrapper{Tuple{},getfield(Main, Symbol("##3#4")){typeof(on_reconf_input),Tuple{InputNode}}}(getfield(Main, Symbol("##3#4")){typeof(on_reconf_input),Tuple{InputNode}}(on_reconf_input, (InputNode(0.0, Publisher{Main.std_msgs.msg.Float64Msg}(PyObject <rospy.topics.Publisher object at 0x7fb506072710>)),)), Dict{Symbol,Any}())>
        # partial(on_reconf_input, node)
        callback = pyfunction(on_reconf_input, Dict{Any,Any}, Int64)
        #server = server_module.Server(InputNodeConfig, callback)
        # spin()
        set_param("value", 1.0)
        loop_rate = Rate(1.0)
        while ! is_shutdown()
            on_reconf_input(node)
            rossleep(loop_rate)
        end
        node, server, callback
    end
end

function on_reconf_input(node::InputNode) # node::InputNode,
    node.input = get_param("value") # config["value"]
    # loginfo(config)
    publish(node.output_topic, std_msgs.msg.Float64Msg(node.input))
    # config
end

function main()
    args = docopt(doc)
    nodes = Dict(
        "pid" => ControlCompute.PID(),
        "proportional" => ControlCompute.Proportional(),
        "integral" => ControlCompute.Integral(),
        "derivative" => ControlCompute.Derivative(),
        "saturate" => ControlCompute.Saturate(),
        "differenciate" => ControlCompute.Differenciate(),
    )
    if !(args["--node"] in keys(nodes) || args["--node"] == "input")#!(ARGS[1] in keys(nodes) || ARGS[1] == "input") #
        logfatal(args["--node"] + " is not a valid node.")
        return
    end

    #if args["--name"] == nothing
    #    init_node("control_node", anonymous=true)
    #else
        init_node("control_node", anonymous=true)
    #end
    if args["--node"] == "input"
        node, server, callback = InputNode()
        # On dirait que la callback est déréférencée dans l'appel par le serveur
        # Il faut investiguer...
        spin()
    else
        node, server = ControlNode(nodes[args["--node"]])
        spin()
    end
end


if ! isinteractive()
    main()
end
