module FindTargets

using Images
"Normalize the coordinates of a point so it does not depends on the
size of the picture.

# Arguments
- `p::Array{Int64,1}`: The point (2-elements)
- `w::Int64`: Width of the picture
- `h::Int64`: height of the picture
# Returns:
[x, y]: The normalized coordinates (2-element Array{Float64,1}).
"
function normalize_coordinates(p, h, w)
    j = max(h,w)
    return [1/j 0; 0 -1/j] * p + [-w/(2*j); h/(2*j)]
end

"""Find three blue targets in the given picture (RGB matrix).

# Arguments:
- `picture`: a 2D matrix of RGB values
- `threshold_blue`: minimal value of the blue channel for a point to be considered as blue.
- `threshold_red`: maximal value of the red channel allowed for a target
- `threshold_green`: maximal value of the green channel allowed for a target
- `return_slices`: Boolean stating if the slices locating the targets should be returned.

# Returns:
    (H,L,R,[objects]) the positions of the targets in the picture (center of mass). objects is the list of slices controlled by the return_slices parameter.

"""
function find_targets(picture; threshold_blue=140, threshold_red=120, threshold_green=190)

    filter_color(c) = (c.r < threshold_red / 255) & (c.g < threshold_green / 255) & (c.b > threshold_blue / 255)

    blue_points = filter_color.(picture)

    structure = [
        false true false
        true  true true
        false true false
    ]
    labels = label_components(blue_points, structure)
    areas = component_lengths(labels)
    areas[1] = 0
    points_indices = sort(1:length(areas),by=x->areas[x], rev=true)
    points_indices = points_indices[1:3]

    centers = component_centroids(labels)
    centers = sort([centers[i] for i in points_indices], by= x -> x[1])
    H = centers[1]
    centers = sort(centers[2:end], by= x -> x[2])
    L = centers[1]
    R = centers[2]
    H, L, R
end

end
